# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import asyncio
import json
import sys
from urllib.parse import urlparse

import aiohttp

from lavacli.__about__ import __version__


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "listen"
    listen_parser = sub.add_parser("listen", help="listen to events")
    listen_parser.add_argument(
        "--filter",
        action="append",
        default=None,
        choices=["device", "event", "testjob", "worker"],
        help="filter by topic type",
    )

    # "wait"
    wait_parser = sub.add_parser("wait", help="wait for a specific event")
    obj_parser = wait_parser.add_subparsers(dest="object", help="object to wait")
    obj_parser.required = True

    # "wait device"
    device_parser = obj_parser.add_parser("device")
    device_parser.add_argument("name", type=str, help="name of the device")
    device_parser.add_argument(
        "--state",
        default=None,
        choices=["IDLE", "RESERVED", "RUNNING"],
        help="device state",
    )
    device_parser.add_argument(
        "--health",
        default=None,
        choices=["GOOD", "UNKNOWN", "LOOPING", "BAD", "MAINTENANCE", "RETIRED"],
        help="device health",
    )

    # "wait job"
    testjob_parser = obj_parser.add_parser("job")
    testjob_parser.add_argument("job_id", help="job id")
    testjob_parser.add_argument(
        "--state",
        default=None,
        choices=[
            "SUBMITTED",
            "SCHEDULING",
            "SCHEDULED",
            "RUNNING",
            "CANCELING",
            "FINISHED",
        ],
        help="job state",
    )
    testjob_parser.add_argument(
        "--health",
        default=None,
        choices=["UNKNOWN", "COMPLETE", "INCOMPLETE", "CANCELED"],
        help="job health",
    )

    # "wait worker"
    worker_parser = obj_parser.add_parser("worker")
    worker_parser.add_argument("name", type=str, help="worker name")
    worker_parser.add_argument(
        "--state", default=None, choices=["ONLINE", "OFFLINE"], help="worker state"
    )
    worker_parser.add_argument(
        "--health",
        default=None,
        choices=["ACTIVE", "MAINTENANCE", "RETIRED"],
        help="worker health",
    )


def help_string():
    return "listen to events"


def _get_zmq_url(proxy, options, config):
    if config is None or config.get("events", {}).get("uri") is None:
        url = proxy.scheduler.get_publisher_event_socket()
        if "*" in url:
            domain = urlparse(options.uri).netloc
            if "@" in domain:
                domain = domain.split("@")[1]
            domain = domain.split(":")[0]
            url = url.replace("*", domain)
    else:
        url = config["events"]["uri"]

    return url


def print_event(options, config, topic, dt, username, data):
    # If unknown, print the full data
    msg = data
    data = json.loads(data)
    # Print according to the topic
    topic_end = topic.split(".")[-1]

    # filter by topic_end
    if options.filter and topic_end not in options.filter:
        return

    if topic_end == "device":
        msg = f"[{data['device']}] <{data['device_type']}> state={data['state']} health={data['health']}"
        if "job" in data:
            msg += " for %s" % data["job"]
    elif topic_end == "testjob":
        msg = f"[{data['job']}] <{data.get('device', '??')}> state={data['state']} health={data['health']} ({data['description']})"
    elif topic_end == "worker":
        msg = f"[{data['hostname']}] state={data['state']} health={data['health']}"
    elif topic_end == "event":
        msg = f"[{data['job']}] message={data['message']}"

    if sys.stdout.isatty():
        print(
            f"\033[1;30m{dt}\033[0m \033[1;37m{topic}\033[0m \033[32m{username}\033[0m - {msg}"
        )
    else:
        print(f"{dt} {topic} {username} - {msg}")


def loop_zmq_events(proxy, options, config, func):
    import zmq  # pylint: disable=import-outside-toplevel
    from zmq.utils.strtypes import b, u  # pylint: disable=import-outside-toplevel

    # Try to find the socket url
    url = _get_zmq_url(proxy, options, config)
    if url is None:
        print("Unable to find the socket url", file=sys.stderr)
        return 1

    context = zmq.Context()
    sock = context.socket(zmq.SUB)
    sock.setsockopt(zmq.SUBSCRIBE, b"")
    # Set the sock proxy (if needed)
    socks = config.get("events", {}).get("socks_proxy")
    if socks is not None:
        print(f"Listening to {url} (socks {socks})")
        sock.setsockopt(zmq.SOCKS_PROXY, b(socks))
    else:
        print("Listening to %s" % url)

    try:
        sock.connect(url)
    except zmq.error.ZMQError as exc:
        print("Unable to connect: %s" % exc, file=sys.stderr)
        return 1

    while True:
        msg = sock.recv_multipart()
        try:
            (topic, _, dt, username, data) = (u(m) for m in msg)
        except ValueError:
            print("Invalid message: %s" % msg, file=sys.stderr)
            continue
        if func(options, config, topic, dt, username, data):
            break


def loop_ws_events(proxy, options, config, func):
    async def handler():
        HEADERS = {"User-Agent": "lavacli v%s" % __version__}
        url = urlparse(options.uri)
        scheme = url.scheme
        path = url.path
        if path.endswith("/RPC2/"):
            path = path[:-5]
        elif path.endswith("/RPC2"):
            path = path[:-4]
        if not path.endswith("/"):
            path = path + "/"

        ws_url = ws_url_redacted = f"{scheme}://{url.netloc}{path}ws/"
        if "@" in url.netloc:
            ws_url_redacted = (
                f"{scheme}://<USERNAME>:<TOKEN>@{url.netloc.split('@')[-1]}{path}ws/"
            )

        try:
            while True:
                try:
                    async with aiohttp.ClientSession(headers=HEADERS) as session:
                        print("Connecting to %s" % ws_url_redacted)
                        async with session.ws_connect(ws_url, heartbeat=30) as ws:
                            async for msg in ws:
                                if msg.type == aiohttp.WSMsgType.TEXT:
                                    try:
                                        data = json.loads(msg.data)
                                        if "error" in data:
                                            raise aiohttp.ClientError(data["error"])
                                        (topic, _, dt, username, data) = data
                                    except ValueError:
                                        print(
                                            "Invalid message: %s" % msg,
                                            file=sys.stderr,
                                        )
                                        continue
                                    if func(options, config, topic, dt, username, data):
                                        return 0
                            print("Connection closed")
                            await asyncio.sleep(1)
                except aiohttp.ClientError as exc:
                    print("Connection issue: %s" % str(exc), file=sys.stderr)
                    await asyncio.sleep(1)
        except Exception as exc:
            print(exc, file=sys.stderr)

    asyncio.run(handler())


def handle_listen(proxy, options, config):
    if config["version"] < (2020, 9):
        loop_zmq_events(proxy, options, config, print_event)
    else:
        loop_ws_events(proxy, options, config, print_event)
    return 0


def handle_wait(proxy, options, config):
    def wait(options, config, topic, dt, username, data):
        # "job" is called "testjob" in the events
        object_topic = options.object
        if object_topic == "job":
            object_topic = "testjob"

        data = json.loads(data)

        # Filter by object
        obj = topic.split(".")[-1]
        if obj != object_topic:
            return False

        if object_topic == "device":
            if data.get("device") != options.name:
                return False
        elif object_topic == "testjob":
            if data.get("job") != options.job_id:
                return False
        else:
            if data.get("hostname") != options.name:
                return False

        # Filter by state
        if options.state is not None:
            if data.get("state") != options.state.capitalize():
                return False
        # Filter by health
        if options.health is not None:
            if data.get("health") != options.health.capitalize():
                return False

        return True

    if config["version"] < (2020, 9):
        loop_zmq_events(proxy, options, config, wait)
    else:
        loop_ws_events(proxy, options, config, wait)
    return 0


def handle(proxy, options, config):
    handlers = {"listen": handle_listen, "wait": handle_wait}
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys
import xmlrpc.client

from lavacli import main


def test_dt_add(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "device-types", "add", "mydt"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.add",
                "args": ("mydt", None, True, False, 24, "hours"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_add_1(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "device-types",
            "add",
            "mydt",
            "--description",
            "my new dt",
            "--hide",
            "--owners-only",
            "--health-frequency",
            "12",
            "--health-denominator",
            "jobs",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.add",
                "args": ("mydt", "my new dt", False, True, 12, "jobs"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_aliases_add(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "aliases", "add", "mydt", "myalias"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.aliases.add",
                "args": ("mydt", "myalias"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_aliases_delete(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "aliases", "delete", "mydt", "myalias"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.aliases.delete",
                "args": ("mydt", "myalias"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_aliases_list(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "aliases", "list", "mydt"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.aliases.list",
                "args": ("mydt",),
                "ret": ["first alias", "second alias"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "Aliases:\n* first alias\n* second alias\n"
    )


def test_dt_aliases_list_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "aliases", "list", "mydt", "--json"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.aliases.list",
                "args": ("mydt",),
                "ret": ["first alias", "second alias"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == '["first alias", "second alias"]\n'  # nosec


def test_dt_aliases_list_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "aliases", "list", "mydt", "--yaml"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.aliases.list",
                "args": ("mydt",),
                "ret": ["first alias", "second alias"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "[first alias, second alias]\n"  # nosec


def test_dt_aliases_list_empty(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "aliases", "list", "mydt"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.aliases.list",
                "args": ("mydt",),
                "ret": [],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Aliases:\n"  # nosec


def test_dt_hc_delete(setup, monkeypatch, capsys, tmp_path):
    version = "2022.4"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "device-types",
            "health-check",
            "delete",
            "mydt",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.set_health_check",
                "args": ("mydt", ""),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_hc_get_after_2019_1(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "health-check", "get", "mydt"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.get_health_check",
                "args": ("mydt",),
                "ret": "definition.yaml",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "definition.yaml\n"  # nosec


def test_dt_hc_set_after_2019_1(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "hc.yaml").open("w") as f_hc:
        f_hc.write("definition")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "device-types",
            "health-check",
            "set",
            "mydt",
            str(tmp_path / "hc.yaml"),
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.set_health_check",
                "args": ("mydt", "definition"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_list(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "device-types", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.list",
                "args": (False,),
                "ret": [
                    {"name": "bbb", "devices": 0, "installed": True, "template": True},
                    {"name": "qemu", "devices": 3, "installed": True, "template": True},
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Device-Types:\n* bbb (0)\n* qemu (3)\n"  # nosec


def test_dt_list_all(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "device-types", "list", "--all"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.list",
                "args": (True,),
                "ret": [
                    {"name": "bbb", "devices": 0, "installed": False, "template": True},
                    {"name": "qemu", "devices": 3, "installed": True, "template": True},
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Device-Types:\n* bbb (0)\n* qemu (3)\n"  # nosec


def test_dt_list_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "device-types", "list", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.list",
                "args": (False,),
                "ret": [
                    {"name": "bbb", "devices": 0, "installed": True, "template": True},
                    {"name": "qemu", "devices": 3, "installed": True, "template": True},
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == [  # nosec
        {"name": "bbb", "devices": 0, "installed": True, "template": True},
        {"name": "qemu", "devices": 3, "installed": True, "template": True},
    ]


def test_dt_list_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "device-types", "list", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.list",
                "args": (False,),
                "ret": [
                    {"name": "bbb", "devices": 0, "installed": True, "template": True},
                    {"name": "qemu", "devices": 3, "installed": True, "template": True},
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {devices: 0, installed: true, name: bbb, template: true}
- {devices: 3, installed: true, name: qemu, template: true}
"""
    )


def test_dt_show(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "device-types", "show", "qemu"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.show",
                "args": ("qemu",),
                "ret": {
                    "name": "qemu",
                    "description": None,
                    "display": True,
                    "owners_only": False,
                    "health_disabled": False,
                    "aliases": ["kvm"],
                    "devices": ["qemu01", "qemu02"],
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """name            : qemu
description     : None
display         : True
owners only     : False
health disabled : False
aliases         : ['kvm']
devices         : ['qemu01', 'qemu02']
"""
    )


def test_dt_show_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "show", "qemu", "--json"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.show",
                "args": ("qemu",),
                "ret": {
                    "name": "qemu",
                    "description": None,
                    "display": True,
                    "owners_only": False,
                    "health_disabled": False,
                    "aliases": ["kvm"],
                    "devices": ["qemu01", "qemu02"],
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == {  # nosec
        "name": "qemu",
        "description": None,
        "display": True,
        "owners_only": False,
        "health_disabled": False,
        "aliases": ["kvm"],
        "devices": ["qemu01", "qemu02"],
    }


def test_dt_show_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "show", "qemu", "--yaml"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.show",
                "args": ("qemu",),
                "ret": {
                    "name": "qemu",
                    "description": None,
                    "display": True,
                    "owners_only": False,
                    "health_disabled": False,
                    "aliases": ["kvm"],
                    "devices": ["qemu01", "qemu02"],
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """aliases: [kvm]
description: null
devices: [qemu01, qemu02]
display: true
health_disabled: false
name: qemu
owners_only: false
"""
    )


def test_dt_template_delete(setup, monkeypatch, capsys, tmp_path):
    version = "2022.4"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "device-types",
            "template",
            "delete",
            "bbb",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.set_template",
                "args": ("bbb", ""),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_template_get(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "template", "get", "bbb"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.get_template",
                "args": ("bbb",),
                "ret": "template content",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "template content\n"  # nosec


def test_dt_template_set(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "template.jinja2").open("w") as f_hc:
        f_hc.write("template definition")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "device-types",
            "template",
            "set",
            "bbb",
            str(tmp_path / "template.jinja2"),
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.set_template",
                "args": ("bbb", "template definition"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_update(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "device-types", "update", "bbb", "--description", "hello"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.update",
                "args": ("bbb", "hello", None, None, None, None, None),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_update_1(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "device-types", "update", "bbb", "--hide"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.update",
                "args": ("bbb", None, False, None, None, None, None),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_update_2(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "device-types",
            "update",
            "bbb",
            "--public",
            "--health-denominator",
            "hours",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.update",
                "args": ("bbb", None, None, False, None, "hours", None),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_dt_update_3(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "device-types",
            "update",
            "bbb",
            "--health-frequency",
            "12",
            "--health-active",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.device_types.update",
                "args": ("bbb", None, None, None, 12, None, False),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import sys

from lavacli import main
from lavacli.utils import safe_yaml


def test_identities_add(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "identities",
            "add",
            "v.l.o",
            "--uri",
            "https://validation.linaro.org/RPC2",
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec

    with (tmp_path / "lavacli.yaml").open() as f_in:
        data = safe_yaml.load(f_in)
        assert set(data.keys()) == {"default", "v.l.o"}  # nosec
        assert data["default"] == {"uri": "https://lava.example.com/RPC2"}  # nosec
        assert data["v.l.o"] == {"uri": "https://validation.linaro.org/RPC2"}  # nosec


def test_identities_add_1(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "identities",
            "add",
            "v.l.o",
            "--uri",
            "https://validation.linaro.org/RPC2",
            "--proxy",
            "http://proxy:3128",
            "--username",
            "admin",
            "--token",
            "12345",
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec

    with (tmp_path / "lavacli.yaml").open() as f_in:
        data = safe_yaml.load(f_in)
        assert set(data.keys()) == {"default", "v.l.o"}  # nosec
        assert data["default"] == {"uri": "https://lava.example.com/RPC2"}  # nosec
        assert data["v.l.o"] == {  # nosec
            "uri": "https://validation.linaro.org/RPC2",
            "proxy": "http://proxy:3128",
            "username": "admin",
            "token": "12345",
        }


def test_identities_add_empty_config(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "identities",
            "add",
            "v.l.o",
            "--uri",
            "https://validation.linaro.org/RPC2",
        ],
    )
    (tmp_path / "lavacli.yaml").unlink()
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec

    with (tmp_path / "lavacli.yaml").open() as f_in:
        data = safe_yaml.load(f_in)
        assert set(data.keys()) == {"v.l.o"}  # nosec
        assert data["v.l.o"] == {"uri": "https://validation.linaro.org/RPC2"}  # nosec


def test_identities_delete(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "delete", "v.l.o"])
    with (tmp_path / "lavacli.yaml").open("w") as f_conf:
        f_conf.write(
            "default:\n  uri: https://lava.example.com/RPC2\nv.l.o:\n  uri: https://validation.linaro.org/RPC"
        )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec

    with (tmp_path / "lavacli.yaml").open() as f_in:
        data = safe_yaml.load(f_in)
        assert list(data.keys()) == ["default"]  # nosec
        assert data["default"] == {"uri": "https://lava.example.com/RPC2"}  # nosec


def test_identities_delete_empty_config(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "delete", "v.l.o"])
    (tmp_path / "lavacli.yaml").unlink()
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unknown identity 'v.l.o'\n"  # nosec
    assert not (tmp_path / "lavacli.yaml").exists()  # nosec


def test_identities_delete_missing_key(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "delete", "v.l.o"])
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unknown identity 'v.l.o'\n"  # nosec

    with (tmp_path / "lavacli.yaml").open() as f_in:
        data = safe_yaml.load(f_in)
        assert list(data.keys()) == ["default"]  # nosec
        assert data["default"] == {"uri": "https://lava.example.com/RPC2"}  # nosec


def test_identities_list(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "list"])
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Identities:\n* default\n"  # nosec

    with (tmp_path / "lavacli.yaml").open() as f_in:
        data = safe_yaml.load(f_in)
        assert list(data.keys()) == ["default"]  # nosec
        assert data["default"] == {"uri": "https://lava.example.com/RPC2"}  # nosec


def test_identities_list_1(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "list"])
    with (tmp_path / "lavacli.yaml").open("w") as f_conf:
        f_conf.write(
            "default:\n  uri: https://lava.example.com/RPC2\nv.l.o:\n  uri: https://validation.linaro.org/RPC2"
        )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Identities:\n* default\n* v.l.o\n"  # nosec

    with (tmp_path / "lavacli.yaml").open() as f_in:
        data = safe_yaml.load(f_in)
        assert set(data.keys()) == {"default", "v.l.o"}  # nosec
        assert data["default"] == {"uri": "https://lava.example.com/RPC2"}  # nosec
        assert data["v.l.o"] == {"uri": "https://validation.linaro.org/RPC2"}  # nosec


def test_identities_list_2(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "list"])
    (tmp_path / "lavacli.yaml").unlink()
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Identities:\n"  # nosec
    assert not (tmp_path / "lavacli.yaml").exists()  # nosec


def test_identities_show(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "show", "default"])
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "uri: https://lava.example.com/RPC2\n"  # nosec


def test_identities_show_missing(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "show", "missing"])
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unknown identity 'missing'\n"  # nosec


def test_identities_show_no_config(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "show", "default"])
    (tmp_path / "lavacli.yaml").unlink()
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unknown identity 'default'\n"  # nosec


def test_identities_show_invalid_config(setup, monkeypatch, capsys, tmp_path):
    monkeypatch.setattr(sys, "argv", ["lavacli", "identities", "show", "default"])
    with (tmp_path / "lavacli.yaml").open("w") as f_conf:
        f_conf.write("hello")
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Invalid configuration file\n"  # nosec



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys

from lavacli.utils import safe_yaml


def configure_parser(parser, _):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "add"
    tags_add = sub.add_parser("add", help="add a tag")
    tags_add.add_argument("tag", help="tag name")
    tags_add.add_argument("--description", default=None, help="tag description")

    # "delete"
    tags_delete = sub.add_parser("delete", help="delete a tag")
    tags_delete.add_argument("tag", help="tag name")

    # "list"
    tags_list = sub.add_parser("list", help="list tags")
    out_format = tags_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "show"
    tags_show = sub.add_parser("show", help="show tag details")
    tags_show.add_argument("tag", help="tag name")
    out_format = tags_show.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        help="print as yaml",
    )


def help_string():
    return "manage device tags"


def handle_add(proxy, options):
    proxy.scheduler.tags.add(options.tag, options.description)
    return 0


def handle_delete(proxy, options):
    proxy.scheduler.tags.delete(options.tag)
    return 0


def handle_list(proxy, options):
    tags = proxy.scheduler.tags.list()
    if options.output_format == "json":
        print(json.dumps(tags))
    elif options.output_format == "yaml":
        safe_yaml.dump(tags, sys.stdout)
    else:
        print("Tags:")
        for tag in tags:
            if tag["description"]:
                print(f"* {tag['name']} ({tag['description']})")
            else:
                print(f"* {tag['name']}")
    return 0


def handle_show(proxy, options):
    tag = proxy.scheduler.tags.show(options.tag)
    if options.output_format == "json":
        print(json.dumps(tag))
    elif options.output_format == "yaml":
        safe_yaml.dump(tag, sys.stdout)
    else:
        print("name       : %s" % tag["name"])
        print("description: %s" % tag["description"])
        print("devices    :")
        for device in tag["devices"]:
            print("* %s" % device)
    return 0


def handle(proxy, options, _):
    handlers = {
        "add": handle_add,
        "delete": handle_delete,
        "list": handle_list,
        "show": handle_show,
    }
    return handlers[options.sub_sub_command](proxy, options)



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import argparse
import os
import socket
import sys
import xml
import xmlrpc.client
from urllib.parse import urlparse

import requests
from requests.adapters import HTTPAdapter
from requests.packages import urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from requests.packages.urllib3.util.retry import Retry

from .__about__ import __version__
from .commands import (
    aliases,
    device_types,
    devices,
    events,
    groups,
    identities,
    jobs,
    lab,
    results,
    system,
    tags,
    users,
    utils,
    workers,
)
from .utils import VERSION_LATEST, exc2str, parse_version, safe_yaml


class RequestsTransport(xmlrpc.client.Transport):
    def __init__(self, scheme, proxy=None, timeout=20.0, verify_ssl_cert=True):
        super().__init__()
        self.scheme = scheme

        # Create a session
        self.session = requests.Session()
        if urllib3.__version__ >= "1.26":
            allowed_methods = "allowed_methods"
        else:
            allowed_methods = "method_whitelist"  # pragma: no cover
        retry_strategy = Retry(
            total=3,
            status_forcelist=[413, 429, 500, 502, 503, 504],
            backoff_factor=1,
            **{allowed_methods: ["HEAD", "OPTIONS", "GET", "POST"]},
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

        # Set the user agent
        self.user_agent = "lavacli v%s" % __version__
        if proxy is None:
            self.proxies = {}
        else:
            self.proxies = {scheme: proxy}
        self.timeout = timeout
        self.verify_ssl_cert = verify_ssl_cert
        if not verify_ssl_cert:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

    def request(self, host, handler, request_body, verbose=False):
        headers = {
            "User-Agent": self.user_agent,
            "Content-Type": "text/xml",
            "Accept-Encoding": "gzip",
        }
        url = f"{self.scheme}://{host}{handler}"
        try:
            response = None
            response = self.session.post(
                url,
                data=request_body,
                headers=headers,
                timeout=self.timeout,
                verify=self.verify_ssl_cert,
                proxies=self.proxies,
            )
            response.raise_for_status()
            return self.parse_response(response)
        except requests.RequestException as e:
            if response is None:
                raise xmlrpc.client.ProtocolError(url, 500, str(e), "")
            else:
                raise xmlrpc.client.ProtocolError(
                    url, response.status_code, str(e), response.headers
                )
        except xml.parsers.expat.ExpatError as e:
            raise xmlrpc.client.ProtocolError(
                url, 500, f"Invalid response format, check your endpoint URL: {e}", ""
            )

    def parse_response(self, response):
        """
        Parse the xmlrpc response.
        """
        p, u = self.getparser()
        p.feed(response.text)
        p.close()
        return u.close()


def load_config(identity):
    # Build the path to the configuration file
    config_dir = os.environ.get("XDG_CONFIG_HOME", "~/.config")
    config_filename = os.path.expanduser(os.path.join(config_dir, "lavacli.yaml"))

    try:
        with open(config_filename, encoding="utf-8") as f_conf:
            config = safe_yaml.load(f_conf.read())
        return config[identity]
    except (FileNotFoundError, KeyError, TypeError):
        return {}


def common_parser():
    parser_obj = argparse.ArgumentParser(add_help=False)

    # --help and --version
    misc = parser_obj.add_argument_group("lavacli")

    misc.add_argument(
        "--help",
        "-h",
        action="store_true",
        default=False,
        help="show this help message and exit",
    )
    misc.add_argument(
        "--version",
        action="store_true",
        default=False,
        help="print the version number and exit",
    )

    # identity or url
    url = parser_obj.add_argument_group("identity").add_mutually_exclusive_group()
    url.add_argument(
        "--uri", type=str, default=None, help="URI of the lava-server RPC endpoint"
    )
    url.add_argument(
        "--identity",
        "-i",
        metavar="ID",
        type=str,
        default="default",
        help="identity stored in the configuration",
    )

    return parser_obj


def parser(parser_obj, commands, version):
    # The sub commands
    root = parser_obj.add_subparsers(dest="sub_command", help="Sub commands")

    keys = list(commands.keys())
    keys.sort()
    for name in keys:
        cls = commands[name]
        cls.configure_parser(root.add_parser(name, help=cls.help_string()), version)

    return parser_obj


def main():
    # List of known commands
    commands = {
        "aliases": aliases,
        "devices": devices,
        "device-types": device_types,
        "events": events,
        "groups": groups,
        "identities": identities,
        "jobs": jobs,
        "lab": lab,
        "results": results,
        "system": system,
        "tags": tags,
        "users": users,
        "utils": utils,
        "workers": workers,
    }

    # Parsing is made of two phases as arguments depends on the API version of
    # the remote server.
    # 1/ Parse the common arguments
    parser_obj = common_parser()
    (options, remaining) = parser_obj.parse_known_args()

    # Do we have to print the version number?
    if options.version:
        print("lavacli %s" % __version__)
        return 0

    # Print help if lavacli is called without any arguments
    if not remaining:
        parser_obj = parser(parser_obj, commands, VERSION_LATEST)
        parser_obj.print_help()
        return 0 if options.help else 1

    # Load the configuration (if any)
    uri = options.uri
    proxy = None
    version = VERSION_LATEST
    config = {}

    lab_validate = False
    if len(remaining) > 2 and remaining[0] == "lab" and remaining[1] == "validate":
        lab_validate = True
    # Skip when not needed.
    if remaining[0] not in ["identities", "utils"] and not lab_validate:
        if uri is None:
            config = load_config(options.identity)
            if config.get("uri") is None:
                print("Unknown identity '%s'" % options.identity, file=sys.stderr)
                return 1
            username = config.get("username")
            token = config.get("token")
            if username is not None and token is not None:
                p = urlparse(config["uri"])
                path = "" if p.path == "/" else p.path
                uri = f"{p.scheme}://{username}:{token}@{p.netloc}{path}"
            else:
                p = urlparse(config["uri"])
                path = "" if p.path == "/" else p.path
                uri = f"{p.scheme}://{p.netloc}{path}"
        else:
            p = urlparse(uri)
            path = "" if p.path == "/" else p.path
            uri = f"{p.scheme}://{p.netloc}{path}"

        try:
            # Create the Transport object
            parsed_uri = urlparse(uri)
            transport = RequestsTransport(
                parsed_uri.scheme,
                config.get("proxy"),
                config.get("timeout", 20.0),
                config.get("verify_ssl_cert", True),
            )
            # allow_none is True because the server does support it
            proxy = xmlrpc.client.ServerProxy(uri, allow_none=True, transport=transport)
            version = proxy.system.version()
        except (OSError, xmlrpc.client.Error) as exc:
            print(f"Unable to connect: {exc2str(exc, options.uri)}", file=sys.stderr)
            return 1

        # Parse version
        version = parse_version(version)
    config["version"] = version

    # Parse the command line
    parser_obj = parser(parser_obj, commands, version)
    options = parser_obj.parse_args()
    options.uri = uri

    try:
        # Run the command
        return commands[options.sub_command].handle(proxy, options, config)
    except (ConnectionError, socket.gaierror) as exc:
        print(f"Unable to connect to '{options.uri}': {exc}", file=sys.stderr)
        return 1

    except KeyboardInterrupt:
        pass

    except identities.ConfigurationError as exc:
        print(exc, file=sys.stderr)
        return 1
    except xmlrpc.client.Error as exc:
        if "sub_sub_command" in options:
            print(
                f"Unable to call '{options.sub_command}.{options.sub_sub_command}': {exc2str(exc, options.uri)}",
                file=sys.stderr,
            )
        else:
            print(
                f"Unable to call '{options.sub_command}': {exc2str(exc, options.uri)}",
                file=sys.stderr,
            )
    except BaseException as exc:
        print("Unknown error: %s" % str(exc), file=sys.stderr)

    return 1



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import sys
import xmlrpc.client

from lavacli import main


def test_users_add_active_staff_superuser(setup, monkeypatch, capsys):
    version = "2023.03"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "users",
            "add",
            "--staff",
            "--superuser",
            "user1",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "auth.users.add",
                "args": ("user1", None, None, None, True, True, True, False),
                "ret": None,
            },
        ],
    )
    assert main() == 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""


def test_users_add_no_active(setup, monkeypatch, capsys):
    version = "2023.03"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "users",
            "add",
            "--no-active",
            "user1",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "auth.users.add",
                "args": ("user1", None, None, None, False, False, False, False),
                "ret": None,
            },
        ],
    )
    assert main() == 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""


def test_users_add_ldap(setup, monkeypatch, capsys):
    version = "2023.03"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "users",
            "add",
            "--ldap",
            "user1",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "auth.users.add",
                "args": ("user1", None, None, None, True, False, False, True),
                "ret": None,
            },
        ],
    )
    assert main() == 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""


def test_users_update_info(setup, monkeypatch, capsys):
    version = "2023.03"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "users",
            "update",
            "--first-name",
            "first",
            "--last-name",
            "last",
            "--email",
            "user1@email.io",
            "user1",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "auth.users.update",
                "args": ("user1", "first", "last", "user1@email.io", None, None, None),
                "ret": None,
            },
        ],
    )
    assert main() == 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""


def test_users_update_active_staff_superuser(setup, monkeypatch, capsys):
    version = "2023.03"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "users",
            "update",
            "--active",
            "--staff",
            "--superuser",
            "user1",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "auth.users.update",
                "args": ("user1", None, None, None, True, True, True),
                "ret": None,
            },
        ],
    )
    assert main() == 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""


def test_users_update_no_active_no_staff_no_superuser(setup, monkeypatch, capsys):
    version = "2023.03"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "users",
            "update",
            "--no-active",
            "--no-staff",
            "--no-superuser",
            "user1",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "auth.users.update",
                "args": ("user1", None, None, None, False, False, False),
                "ret": None,
            },
        ],
    )
    assert main() == 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys

from lavacli.utils import safe_yaml


def configure_parser(parser, _):
    parser.add_argument("job_id", help="job id")
    parser.add_argument("test_suite", nargs="?", default=None, help="test suite")
    parser.add_argument("test_case", nargs="?", default=None, help="test case")
    out_format = parser.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        default=None,
        action="store_const",
        const="yaml",
        help="print as yaml",
    )


def help_string():
    return "manage results"


def handle(proxy, options, _):
    if options.test_case is not None:
        data = proxy.results.get_testcase_results_yaml(
            options.job_id, options.test_suite, options.test_case
        )
    elif options.test_suite is not None:
        data = proxy.results.get_testsuite_results_yaml(
            options.job_id, options.test_suite
        )
    else:
        data = proxy.results.get_testjob_results_yaml(options.job_id)

    results = safe_yaml.load(data)

    if options.output_format == "json":
        print(json.dumps(results))
    elif options.output_format == "yaml":
        safe_yaml.dump(results, sys.stdout)
    else:
        # Only print the result
        if options.test_case is not None:
            for res in results:
                if not sys.stdout.isatty():
                    print("%s" % res["result"])
                elif res["result"] == "pass":
                    print("\033[1;32mpass\033[0m")
                elif res["result"] == "fail":
                    print("\033[1;31mfail\033[0m")
                else:
                    print("%s" % res["result"])
        # A list to print
        else:
            print("Results:")
            for res in results:
                if not sys.stdout.isatty():
                    print(f"* {res['suite']}.{res['name']} [{res['result']}]")
                elif res["result"] == "pass":
                    print(f"* {res['suite']}.{res['name']} [\033[1;32mpass\033[0m]")
                elif res["result"] == "fail":
                    print(f"* {res['suite']}.{res['name']} [\033[1;31mfail\033[0m]")
                else:
                    print(f"* {res['suite']}.{res['name']} [{res['result']}]")
    return 0



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import argparse
import json
import sys
import time

from lavacli.utils import safe_yaml


def positive_integer(value):
    int_value = int(value)
    if int_value < 0:
        raise argparse.ArgumentTypeError("%r should be a positive integer" % value)
    return int_value


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "add"
    workers_add = sub.add_parser("add", help="add a worker")
    workers_add.add_argument("name", type=str, help="worker name")
    workers_add.add_argument(
        "--description", type=str, default=None, help="worker description"
    )
    workers_add.add_argument(
        "--disabled",
        action="store_true",
        default=False,
        help="create a disabled worker",
    )

    # "config"
    workers_config = sub.add_parser("config", help="worker configuration")
    config_sub = workers_config.add_subparsers(
        dest="sub_sub_sub_command", help="Sub commands"
    )
    config_sub.required = True
    if version >= (2022, 4):
        config_delete = config_sub.add_parser(
            "delete", help="delete the worker configuration"
        )
        config_delete.add_argument("name", type=str, help="worker name")
    config_get = config_sub.add_parser("get", help="get the worker configuration")
    config_get.add_argument("name", type=str, help="worker name")

    config_set = config_sub.add_parser("set", help="set the worker configuration")
    config_set.add_argument("name", type=str, help="worker name")
    config_set.add_argument(
        "config", type=argparse.FileType("r"), help="configuration file"
    )

    if (2020, 4) <= version <= (2020, 8):
        # "certificate"
        workers_cert = sub.add_parser("certificate", help="worker certificate")
        cert_sub = workers_cert.add_subparsers(
            dest="sub_sub_sub_command", help="Sub commands"
        )
        cert_sub.required = True
        cert_get = cert_sub.add_parser("get", help="get the worker certificate")
        cert_get.add_argument("name", type=str, help="worker name")

        cert_set = cert_sub.add_parser("set", help="set the worker certificate")
        cert_set.add_argument("name", type=str, help="worker name")
        cert_set.add_argument(
            "certificate", type=argparse.FileType("r"), help="certificate file"
        )

    if version >= (2019, 6):
        # "env"
        workers_env = sub.add_parser("env", help="worker environment")
        env_sub = workers_env.add_subparsers(
            dest="sub_sub_sub_command", help="Sub commands"
        )
        env_sub.required = True
        if version >= (2022, 4):
            env_delete = env_sub.add_parser(
                "delete", help="delete the worker environment"
            )
            env_delete.add_argument("name", type=str, help="worker name")

        env_get = env_sub.add_parser("get", help="get the worker environment")
        env_get.add_argument("name", type=str, help="worker name")

        env_set = env_sub.add_parser("set", help="set the worker environment")
        env_set.add_argument("name", type=str, help="worker name")
        env_set.add_argument(
            "env", type=argparse.FileType("r"), help="environment file"
        )

    if version >= (2022, 1):
        # "env-dut"
        workers_env_dut = sub.add_parser("env-dut", help="worker environment")
        env_dut_sub = workers_env_dut.add_subparsers(
            dest="sub_sub_sub_command", help="Sub commands"
        )
        env_dut_sub.required = True
        if version >= (2022, 4):
            env_dut_delete = env_dut_sub.add_parser(
                "delete", help="delete the dut environment"
            )
            env_dut_delete.add_argument("name", type=str, help="worker name")

        env_dut_get = env_dut_sub.add_parser("get", help="get the dut environment")
        env_dut_get.add_argument("name", type=str, help="worker name")

        env_dut_set = env_dut_sub.add_parser("set", help="set the dut environment")
        env_dut_set.add_argument("name", type=str, help="worker name")
        env_dut_set.add_argument(
            "env", type=argparse.FileType("r"), help="environment file"
        )

    # "list"
    workers_list = sub.add_parser("list", help="list workers")
    workers_list.add_argument(
        "--all",
        "-a",
        action="store_true",
        default=False,
        help="list every workers, including retired",
    )
    out_format = workers_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )

    # "maintenance"
    workers_maintenance = sub.add_parser("maintenance", help="maintenance the worker")
    workers_maintenance.add_argument("name", type=str, help="worker name")
    workers_maintenance.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="force worker maintenance by canceling running jobs",
    )
    workers_maintenance.add_argument(
        "--no-wait",
        dest="wait",
        default=True,
        action="store_false",
        help="do not wait for the devices to be idle",
    )

    # "show"
    workers_show = sub.add_parser("show", help="show worker details")
    workers_show.add_argument("name", help="worker name")
    out_format = workers_show.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )

    # "update"
    update_parser = sub.add_parser("update", help="update worker properties")
    update_parser.add_argument("name", type=str, help="worker name")
    update_parser.add_argument(
        "--description", type=str, default=None, help="worker description"
    )
    update_parser.add_argument(
        "--health",
        type=str,
        default=None,
        choices=["ACTIVE", "MAINTENANCE", "RETIRED"],
        help="worker health",
    )
    if version >= (2020, 1):
        update_parser.add_argument(
            "--job-limit", type=positive_integer, default=None, help="job limit"
        )


def help_string():
    return "manage workers"


def handle_add(proxy, options, _):
    proxy.scheduler.workers.add(options.name, options.description, options.disabled)
    return 0


def handle_config(proxy, options, _):
    if options.sub_sub_sub_command == "delete":
        config = proxy.scheduler.workers.set_config(options.name, "")
    elif options.sub_sub_sub_command == "get":
        config = proxy.scheduler.workers.get_config(options.name)
        print(str(config).rstrip("\n"))
    else:
        config = options.config.read()
        ret = proxy.scheduler.workers.set_config(options.name, config)
        if not ret:
            print("Unable to store worker configuration", file=sys.stderr)
            return 1
    return 0


def handle_certificate(proxy, options, _):
    if options.sub_sub_sub_command == "get":
        env = proxy.scheduler.workers.get_certificate(options.name)
        print(str(env).rstrip("\n"))
    else:
        certificate = options.certificate.read()
        ret = proxy.scheduler.workers.set_certificate(options.name, certificate)
        if not ret:
            print("Unable to store worker certificate", file=sys.stderr)
            return 1
    return 0


def handle_env(proxy, options, _):
    if options.sub_sub_sub_command == "delete":
        proxy.scheduler.workers.set_env(options.name, "")
    elif options.sub_sub_sub_command == "get":
        env = proxy.scheduler.workers.get_env(options.name)
        print(str(env).rstrip("\n"))
    else:
        env = options.env.read()
        ret = proxy.scheduler.workers.set_env(options.name, env)
        if not ret:
            print("Unable to store worker environment", file=sys.stderr)
            return 1
    return 0


def handle_env_dut(proxy, options, _):
    if options.sub_sub_sub_command == "delete":
        proxy.scheduler.workers.set_env_dut(options.name, "")
    elif options.sub_sub_sub_command == "get":
        env = proxy.scheduler.workers.get_env_dut(options.name)
        print(str(env).rstrip("\n"))
    else:
        env = options.env.read()
        ret = proxy.scheduler.workers.set_env_dut(options.name, env)
        if not ret:
            print("Unable to store worker dut environment", file=sys.stderr)
            return 1
    return 0


def handle_list(proxy, options, config):
    if config["version"] >= (2023, 3):
        workers = proxy.scheduler.workers.list(options.all)
    else:
        workers = proxy.scheduler.workers.list()
    if options.output_format == "json":
        print(json.dumps(workers))
    elif options.output_format == "yaml":
        safe_yaml.dump(workers, sys.stdout)
    else:
        print("Workers:")
        for worker in workers:
            print("* %s" % worker)
    return 0


def handle_maintenance(proxy, options, _):
    proxy.scheduler.workers.update(options.name, None, "MAINTENANCE")

    if options.force or options.wait:
        worker_devices = proxy.scheduler.workers.show(options.name)["devices"]
        for device in proxy.scheduler.devices.list():
            if device["hostname"] not in worker_devices:
                continue
            current_job = device["current_job"]
            if current_job is not None:
                print("-> waiting for job %s" % current_job)
                # if --force is passed, cancel the job
                if options.force:
                    print("--> canceling")
                    proxy.scheduler.jobs.cancel(current_job)
                while (
                    options.wait
                    and proxy.scheduler.jobs.show(current_job)["state"] != "Finished"
                ):
                    print("--> waiting")
                    time.sleep(5)
    return 0


def handle_show(proxy, options, config):
    worker = proxy.scheduler.workers.show(options.name)
    if options.output_format == "json":
        if "last_ping" in worker:
            worker["last_ping"] = worker["last_ping"].value
        print(json.dumps(worker))
    elif options.output_format == "yaml":
        if "last_ping" in worker:
            worker["last_ping"] = worker["last_ping"].value
        safe_yaml.dump(worker, sys.stdout)
    else:
        print("hostname    : %s" % worker["hostname"])
        print("description : %s" % worker["description"])
        print("state       : %s" % worker["state"])
        print("health      : %s" % worker["health"])
        print("devices     : %s" % ", ".join(worker["devices"]))
        print("last ping   : %s" % worker["last_ping"])

        if config["version"] >= (2020, 1):
            print("job limit   : %d" % worker["job_limit"])

        if config["version"] >= (2020, 6):
            print("version     : %s" % worker["version"])

        if "token" in worker:
            print("token       : %s" % worker["token"])
    return 0


def handle_update(proxy, options, config):
    if config["version"] >= (2020, 1):
        proxy.scheduler.workers.update(
            options.name, options.description, options.health, options.job_limit
        )
    else:
        proxy.scheduler.workers.update(
            options.name, options.description, options.health
        )
    return 0


def handle(proxy, options, config):
    handlers = {
        "add": handle_add,
        "config": handle_config,
        "certificate": handle_certificate,
        "env": handle_env,
        "env-dut": handle_env_dut,
        "list": handle_list,
        "maintenance": handle_maintenance,
        "show": handle_show,
        "update": handle_update,
    }
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import argparse
import contextlib
import datetime
import json
import pathlib
import sys
import time
import xmlrpc.client
from urllib.parse import urlparse

from lavacli.utils import exc2str, log_yaml, safe_yaml
from lavacli.utils import fetch_jobs, print_error_summary
from collections import defaultdict


def configure_log_options(parser):
    parser.add_argument(
        "--filters",
        default=None,
        type=str,
        help="comma separated list of levels to show",
    )
    parser.add_argument(
        "--polling",
        default=5,
        type=int,
        help="polling interval in seconds [default=5s]",
    )
    parser.add_argument(
        "--raw", default=False, action="store_true", help="print raw logs"
    )
    parser.add_argument("--start", default=0, type=int, help="start at the given line")
    parser.add_argument("--end", default=None, type=int, help="end at the given line")


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "cancel"
    jobs_cancel = sub.add_parser("cancel", help="cancel a job")
    jobs_cancel.add_argument("job_id", nargs="+", help="job id")

    if version >= (2023, 10):
        # "autocancel"
        jobs_autocancel = sub.add_parser(
            "autocancel",
            help="cancel jobs that take longer than job timeout and the specific duration",
        )
        jobs_autocancel.add_argument(
            "--duration", type=int, required=True, help="duration in minutes"
        )
        jobs_autocancel.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="do not cancel the jobs",
        )

    # "config"
    jobs_config = sub.add_parser("config", help="job configuration")
    jobs_config.add_argument("job_id", help="job id")
    jobs_config.add_argument(
        "--dest", default=".", help="save files into this directory"
    )

    # "definition"
    jobs_definition = sub.add_parser("definition", help="job definition")
    jobs_definition.add_argument("job_id", help="job id")

    # "find-errors"
    jobs_find_errors = sub.add_parser("find-errors", help="Find errors in recent jobs")
    jobs_find_errors.add_argument(
        "--limit", type=int, default=100, help="Number of jobs to fetch"
    )
    jobs_find_errors.add_argument(
        "--since", type=int, default=None, help="Jobs from the last N minutes"
    )

    # "list"
    jobs_list = sub.add_parser("list", help="list jobs")
    jobs_list.add_argument(
        "--state",
        type=str,
        default=None,
        choices=[
            "SUBMITTED",
            "SCHEDULING",
            "SCHEDULED",
            "RUNNING",
            "CANCELING",
            "FINISHED",
        ],
        help="filter jobs by state",
    )
    jobs_list.add_argument(
        "--health",
        type=str,
        default=None,
        choices=["UNKNOWN", "COMPLETE", "INCOMPLETE", "CANCELED"],
        help="filter jobs by health",
    )
    jobs_list.add_argument(
        "--since",
        type=int,
        default=0,
        help="Filter by jobs which completed in the last N minutes.",
    )
    jobs_list.add_argument(
        "--verbose",
        dest="verbose",
        action="store_true",
        help="If verbose is True, add extra keys, including error_type.",
    )
    jobs_list.add_argument(
        "--start", type=int, default=0, help="skip the N first jobs [default=0]"
    )
    jobs_list.add_argument(
        "--limit", type=int, default=25, help="limit to N jobs [default=25]"
    )
    out_format = jobs_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        default=None,
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "logs"
    jobs_logs = sub.add_parser("logs", help="get logs")
    jobs_logs.add_argument("job_id", help="job id")
    jobs_logs.add_argument(
        "--no-follow",
        default=False,
        action="store_true",
        help="do not keep polling until the end of the job",
    )
    configure_log_options(jobs_logs)

    # "queue"
    jobs_queue = sub.add_parser("queue", help="job queue")
    jobs_queue.add_argument("device_types", nargs="*", help="filter by device-types")
    out_format = jobs_queue.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )
    jobs_queue.add_argument(
        "--start", type=int, default=0, help="skip the N first jobs [default=0]"
    )
    jobs_queue.add_argument(
        "--limit", type=int, default=25, help="limit to N jobs [default=25]"
    )

    # "resubmit"
    jobs_resubmit = sub.add_parser("resubmit", help="resubmit a job")
    jobs_resubmit.add_argument("job_id", help="job id")
    jobs_resubmit.add_argument(
        "--url",
        action="store_true",
        dest="print_url",
        default=False,
        help="Print the full url",
    )
    jobs_resubmit.add_argument(
        "--follow",
        default=True,
        dest="no_follow",
        action="store_false",
        help="resubmit and poll for the logs",
    )
    configure_log_options(jobs_resubmit)

    # "run"
    jobs_run = sub.add_parser("run", help="run the job")
    jobs_run.add_argument(
        "definition", type=argparse.FileType("r"), help="job definition"
    )
    jobs_run.add_argument(
        "--filters",
        default=None,
        type=str,
        help="comma separated list of levels to show",
    )
    jobs_run.add_argument(
        "--no-follow",
        default=False,
        action="store_true",
        help="do not keep polling until the end of the job",
    )
    jobs_run.add_argument(
        "--polling",
        default=5,
        type=int,
        help="polling interval in seconds [default=5s]",
    )
    jobs_run.add_argument(
        "--raw", default=False, action="store_true", help="print raw logs"
    )
    jobs_run.add_argument(
        "--start", default=0, type=int, help="start at the given line"
    )
    jobs_run.add_argument("--end", default=None, type=int, help="end at the given line")

    # "show"
    jobs_show = sub.add_parser("show", help="job details")
    jobs_show.add_argument("job_id", help="job id")
    out_format = jobs_show.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )

    # "submit"
    jobs_submit = sub.add_parser("submit", help="submit a new job")
    jobs_submit.add_argument(
        "definition", nargs="+", type=argparse.FileType("r"), help="job definition"
    )
    jobs_submit.add_argument(
        "--url",
        action="store_true",
        dest="print_url",
        default=False,
        help="Print the full url",
    )
    jobs_submit.add_argument(
        "--follow",
        default=True,
        dest="no_follow",
        action="store_false",
        help="resubmit and poll for the logs",
    )
    configure_log_options(jobs_submit)

    if version >= (2019, 3):
        # "validate"
        jobs_validate = sub.add_parser("validate", help="validate the job definition")
        jobs_validate.add_argument(
            "definition", type=argparse.FileType("r"), help="job definition"
        )
        jobs_validate.add_argument(
            "--strict", action="store_true", default=False, help="check in strict mode"
        )

    # "wait"
    jobs_wait = sub.add_parser("wait", help="wait for the job to finish")
    jobs_wait.add_argument("job_id", help="job id")
    jobs_wait.add_argument(
        "--polling",
        default=5,
        type=int,
        help="polling interval in seconds [default=5s]",
    )
    jobs_wait.add_argument(
        "--timeout",
        default=0,
        type=int,
        help="Maximum time to wait in seconds, 0 to disable [default=0s]",
    )


def help_string():
    return "manage jobs"


def handle_cancel(proxy, options, _):
    ret = 0
    for job_id in options.job_id:
        try:
            proxy.scheduler.jobs.cancel(job_id)
        except xmlrpc.client.Error as exc:
            print(
                f"Unable to cancel job {job_id}: {exc2str(exc, options.uri)}",
                file=sys.stderr,
            )
            ret = 1
    return ret


def handle_autocancel(proxy, options, _):
    # Jobs that still running within its timeout shouldn't be canceled.
    skip_list = []
    # A loop is needed as the list api only returns the last 100 jobs.
    while True:
        jobs = proxy.scheduler.jobs.list(
            "RUNNING",
            None,
            0,
            100,
            None,
            False,
            options.duration,
        )

        if not jobs:
            break

        job_ids = [job["id"] for job in jobs]
        if set(job_ids).issubset(skip_list):
            break
        jobs = [job for job in jobs if job["id"] not in skip_list]

        duration = datetime.timedelta(minutes=options.duration)
        for job in jobs:
            job_id = job["id"]
            definition = proxy.scheduler.jobs.definition(job_id)
            data = safe_yaml.load(definition)["timeouts"]["job"]
            job_timeout = datetime.timedelta(
                days=data.get("days", 0),
                hours=data.get("hours", 0),
                minutes=data.get("minutes", 0),
                seconds=data.get("seconds", 0),
            )
            if duration > job_timeout:
                try:
                    if not options.dry_run:
                        proxy.scheduler.jobs.cancel(job_id)
                    print(
                        "[CANCELED] %s: [%s] (%s) - %s"
                        % (
                            job_id,
                            job["submitter"],
                            job["description"],
                            job["device_type"],
                        )
                    )
                except xmlrpc.client.Error as exc:
                    print(
                        f"Unable to cancel {job_id}: {exc2str(exc, options.uri)}",
                        file=sys.stderr,
                    )
                    return 1
            else:
                print(
                    "[  SKIP  ] %s: [%s] (%s) - %s - timeout not reached yet"
                    % (
                        job_id,
                        job["submitter"],
                        job["description"],
                        job["device_type"],
                    )
                )
                skip_list.append(job_id)

    return 0


def handle_config(proxy, options, _):
    config = proxy.scheduler.jobs.configuration(options.job_id)
    definition, device, dispatcher, env, env_dut = config

    # Create the directory if needed
    dest = pathlib.Path(options.dest)
    with contextlib.suppress(FileExistsError):
        dest.mkdir(mode=0o755)

    # Save the files
    with (dest / "definition.yaml").open("w") as f_out:
        f_out.write(definition)
    if device is not None:
        with (dest / "device.yaml").open("w") as f_out:
            f_out.write(device)
    if dispatcher is not None:
        with (dest / "dispatcher.yaml").open("w") as f_out:
            f_out.write(dispatcher)
    if env is not None:
        with (dest / "env.yaml").open("w") as f_out:
            f_out.write(env)
    if env_dut is not None:
        with (dest / "env.dut.yaml").open("w") as f_out:
            f_out.write(env_dut)
    return 0


def handle_definition(proxy, options, _):
    print(proxy.scheduler.jobs.definition(options.job_id))
    return 0


def handle_list(proxy, options, config):
    jobs = proxy.scheduler.jobs.list(
        options.state,
        options.health,
        options.start,
        options.limit,
        options.since,
        options.verbose,
    )

    if options.output_format == "json":
        print(json.dumps(jobs))
    elif options.output_format == "yaml":
        safe_yaml.dump(jobs, sys.stdout)
    else:
        print(f"Jobs (from {1 + options.start} to {options.start + options.limit}):")
        for job in jobs:
            if options.verbose:
                if job["error_type"]:
                    print(
                        "* %s: %s,%s [%s] (%s) - %s %s <%s> <%s> %s: %s"
                        % (
                            job["id"],
                            job["state"],
                            job["health"],
                            job["submitter"],
                            job["description"],
                            job["device_type"],
                            job["actual_device"],
                            job["start_time"],
                            job["end_time"],
                            job["error_type"],
                            job["error_msg"],
                        )
                    )
                else:
                    print(
                        "* %s: %s,%s [%s] (%s) - %s %s <%s> <%s>"
                        % (
                            job["id"],
                            job["state"],
                            job["health"],
                            job["submitter"],
                            job["description"],
                            job["device_type"],
                            job["actual_device"],
                            job["start_time"],
                            job["end_time"],
                        )
                    )

            else:
                print(
                    "* %s: %s,%s [%s] (%s) - %s"
                    % (
                        job["id"],
                        job["state"],
                        job["health"],
                        job["submitter"],
                        job["description"],
                        job["device_type"],
                    )
                )
    return 0


if sys.stdout.isatty():
    COLORS = {
        "exception": "\033[1;31m",
        "error": "\033[1;31m",
        "warning": "\033[1;33m",
        "info": "\033[1;37m",
        "debug": "\033[0;37m",
        "target": "\033[32m",
        "input": "\033[0;35m",
        "feedback": "\033[0;33m",
        "results": "\033[1;34m",
        "dt": "\033[0;90m",
        "end": "\033[0m",
    }
else:
    COLORS = {
        "exception": "",
        "error": "",
        "warning": "",
        "info": "",
        "debug": "",
        "target": "",
        "input": "",
        "feedback": "",
        "results": "",
        "dt": "",
        "end": "",
    }


def print_logs(logs, raw, filters):
    filters = [] if filters is None else filters.split(",")
    if raw:
        for line in logs:
            if filters and not line["lvl"] in filters:
                continue
            sys.stdout.write("- ")
            log_yaml.dump(
                line,
                sys.stdout,
            )
    else:
        for line in logs:
            timestamp = line["dt"].split(".")[0]
            level = line["lvl"]
            if filters and level not in filters:
                continue
            if isinstance(line["msg"], dict) and "sending" in line["msg"].keys():
                level = "input"
                msg = str(line["msg"]["sending"])
            elif isinstance(line["msg"], bytes):
                msg = line["msg"].decode("utf-8", errors="replace")
            else:
                msg = str(line["msg"])
            msg = msg.rstrip("\n")

            ns = ""
            if level == "feedback" and "ns" in line:
                ns = " <" + COLORS["feedback"] + line["ns"] + COLORS["end"] + ">"
            print(
                COLORS["dt"]
                + timestamp
                + COLORS["end"]
                + ns
                + " "
                + COLORS[level]
                + msg
                + COLORS["end"]
            )


def _download_logs(proxy, version, job_id, start, end):
    (finished, data) = proxy.scheduler.jobs.logs(job_id, start, end)
    logs = safe_yaml.load(str(data))
    return (finished, logs)


def handle_logs(proxy, options, config):
    # Loop
    lines = options.start
    while True:
        try:
            (finished, logs) = _download_logs(
                proxy, config["version"], options.job_id, lines, options.end
            )
            if logs:
                print_logs(logs, options.raw, options.filters)
                lines += len(logs)
            # Loop only if the job is not finished
            if (
                finished
                or options.no_follow
                or (options.end is not None and lines >= options.end)
            ):
                break
        except xmlrpc.client.ProtocolError as exc:
            # Ignore temporary Internal Server Error as we already got some logs
            if lines > options.start and exc.errcode == 500:
                pass

        # Wait some time
        time.sleep(options.polling)

    # Print the failure comment if the job is finished
    if finished:
        details = proxy.scheduler.jobs.show(options.job_id)
        if details.get("failure_comment"):
            print_logs(
                [
                    {
                        "dt": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                        "lvl": "info",
                        "msg": "[lavacli] Failure comment: %s"
                        % details["failure_comment"],
                    }
                ],
                options.raw,
                options.filters,
            )
    return 0


def handle_queue(proxy, options, config):
    dts = options.device_types if options.device_types else None
    data = proxy.scheduler.jobs.queue(dts, options.start, options.limit)
    if options.output_format == "json":
        print(json.dumps(data))
    elif options.output_format == "yaml":
        safe_yaml.dump(data, sys.stdout)
    else:
        print(f"Jobs (from {1+options.start} to {options.start + options.limit}):")
        for job in data:
            print(
                f"* {job['id']}: {job['submitter']} ({job['description'] or ''}) - {job['requested_device_type']}"
            )
    return 0


def handle_resubmit(proxy, options, config):
    job_id = proxy.scheduler.jobs.resubmit(options.job_id)

    if options.no_follow:
        prefix = ""
        if options.print_url:
            parsed = urlparse(options.uri)
            host = parsed.netloc.split("@")[-1]
            prefix = f"{parsed.scheme}://{host}/scheduler/job/"

        if isinstance(job_id, list):
            for job in job_id:
                print(prefix + str(job))
        else:
            print(prefix + str(job_id))

    else:
        print_logs(
            [
                {
                    "dt": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                    "lvl": "info",
                    "msg": "[lavacli] Job %s submitted" % job_id,
                }
            ],
            options.raw,
            options.filters,
        )

        follow_logs(job_id, proxy, options, config)
    return 0


def follow_logs(job_id, proxy, options, config):
    # Add the job_id to options for handle_logs
    # For multinode, print something and loop on all jobs
    if isinstance(job_id, list):
        for job in job_id:
            print_logs(
                [
                    {
                        "dt": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                        "lvl": "info",
                        "msg": "[lavacli] Seeing %s logs" % job,
                    }
                ],
                options.raw,
                options.filters,
            )
            options.job_id = job
            handle_logs(proxy, options, config)
    else:
        options.job_id = str(job_id)
        handle_logs(proxy, options, config)


def handle_run(proxy, options, config):
    job_id = proxy.scheduler.jobs.submit(options.definition.read())
    print_logs(
        [
            {
                "dt": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                "lvl": "info",
                "msg": "[lavacli] Job %s submitted" % job_id,
            }
        ],
        options.raw,
        options.filters,
    )

    # Add the job_id to options for handle_logs
    # For multinode, print something and loop on all jobs
    if isinstance(job_id, list):
        for job in job_id:
            print_logs(
                [
                    {
                        "dt": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                        "lvl": "info",
                        "msg": "[lavacli] Seeing %s logs" % job,
                    }
                ],
                options.raw,
                options.filters,
            )
            options.job_id = job
            handle_logs(proxy, options, config)
    else:
        options.job_id = job_id
        handle_logs(proxy, options, config)
    return 0


def handle_show(proxy, options, config):
    job = proxy.scheduler.jobs.show(options.job_id)

    if options.output_format == "json":
        job["submit_time"] = job["submit_time"].value or None
        job["start_time"] = job["start_time"].value or None
        job["end_time"] = job["end_time"].value or None
        print(json.dumps(job))
    elif options.output_format == "yaml":
        job["submit_time"] = job["submit_time"].value or None
        job["start_time"] = job["start_time"].value or None
        job["end_time"] = job["end_time"].value or None
        safe_yaml.dump(job, sys.stdout)
    else:
        print("id          : %s" % job["id"])
        print("description : %s" % job["description"])
        print("submitter   : %s" % job["submitter"])
        print("device-type : %s" % job["device_type"])
        print("device      : %s" % job["device"])
        print("health-check: %s" % job["health_check"])
        print("state       : %s" % job["state"])
        print("Health      : %s" % job["health"])
        if job.get("failure_comment"):
            print("failure     : %s" % job["failure_comment"])
        if job.get("error_type"):
            print("error       : %s: %s" % (job["error_type"], job["error_msg"]))
        print("pipeline    : %s" % job["pipeline"])
        print("tags        : %s" % str(job["tags"]))
        print("visibility  : %s" % job["visibility"])
        print("submit time : %s" % job["submit_time"])
        print("start time  : %s" % job["start_time"])
        print("end time    : %s" % job["end_time"])
    return 0


def handle_submit(proxy, options, config):
    prefix = ""
    if options.print_url:
        parsed = urlparse(options.uri)
        host = parsed.netloc.split("@")[-1]
        prefix = f"{parsed.scheme}://{host}/scheduler/job/"

    for definition in options.definition:
        try:
            job_id = proxy.scheduler.jobs.submit(definition.read())
            if options.no_follow:
                if isinstance(job_id, list):
                    for job in job_id:
                        print(prefix + str(job))
                else:
                    print(prefix + str(job_id))
        except xmlrpc.client.Error as exc:
            print(
                f"Unable to submit {definition.name}: {exc2str(exc, options.uri)}",
                file=sys.stderr,
            )
            return 1

    if not options.no_follow:
        follow_logs(job_id, proxy, options, config)

    return 0


def handle_validate(proxy, options, _):
    ret = proxy.scheduler.jobs.validate(options.definition.read(), options.strict)
    if not ret:
        return 0
    print("key: %s" % ret["path"])
    print("msg: %s" % ret["msg"])
    return 1


def handle_wait(proxy, options, _):
    job = proxy.scheduler.jobs.show(options.job_id)
    old_state = ""
    time_elapsed = 0
    while job["state"] != "Finished":
        if old_state != job["state"]:
            if old_state:
                sys.stdout.write("\n")
            sys.stdout.write(job["state"])
        else:
            sys.stdout.write(".")
        sys.stdout.flush()
        old_state = job["state"]
        time.sleep(options.polling)
        job = proxy.scheduler.jobs.show(options.job_id)
        if options.timeout > 0:
            time_elapsed = time_elapsed + options.polling
            if time_elapsed > options.timeout:
                print("timeout!")
                return 1
    if old_state != job["state"] and old_state:
        sys.stdout.write("\n")
    return 0


def handle_find_errors(proxy, options, config):
    """Main function to fetch and process LAVA jobs."""

    parsed = urlparse(options.uri)
    host = parsed.netloc.split("@")[-1]
    url = f"{parsed.scheme}://{host}"

    job_numbers = fetch_jobs(proxy, None, "INCOMPLETE", options.limit, options.since)
    if not job_numbers:
        print("No jobs found.")
        return

    print(f"\nTotal jobs found: {len(job_numbers)}")

    error_summary = defaultdict(lambda: defaultdict(list))

    for job_id in job_numbers:
        try:
            job = proxy.scheduler.jobs.show(job_id)
        except xmlrpc.client.Error as exc:
            print(f"Error fetching job {job_id}: {exc}")
            continue

        device_name = job.get("device") or job.get("device_type")
        job_url = f"{url}/scheduler/job/{job_id}"

        error_msg = job.get("error_msg", "").strip()
        if error_msg:
            error_summary[error_msg][device_name].append(job_url)

    print_error_summary(error_summary, summary=True)
    print_error_summary(error_summary, summary=False)


def handle(proxy, options, config):
    handlers = {
        "cancel": handle_cancel,
        "autocancel": handle_autocancel,
        "config": handle_config,
        "definition": handle_definition,
        "list": handle_list,
        "logs": handle_logs,
        "queue": handle_queue,
        "resubmit": handle_resubmit,
        "run": handle_run,
        "show": handle_show,
        "submit": handle_submit,
        "validate": handle_validate,
        "wait": handle_wait,
        "find-errors": handle_find_errors,
    }
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import os
import sys

from lavacli.utils import flow_yaml, safe_yaml


class ConfigurationError(Exception):
    pass


def configure_parser(parser, _):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    # "add"
    config_add = sub.add_parser("add", help="add an identity")
    config_add.add_argument("id", type=str, help="identity")
    config_add.add_argument(
        "--uri", type=str, required=True, help="URI of the lava-server RPC endpoint"
    )
    config_add.add_argument("--username", type=str, default=None, help="api username")
    config_add.add_argument("--token", type=str, default=None, help="api token")
    config_add.add_argument("--proxy", type=str, default=None, help="http proxy")

    # "delete"
    config_del = sub.add_parser("delete", help="delete an alias")
    config_del.add_argument("id", help="identity")

    # "list"
    sub.add_parser("list", help="list available identities")

    # "show"
    config_show = sub.add_parser("show", help="show identity details")
    config_show.add_argument("id", type=str, help="identity")


def help_string():
    return "manage lavacli configuration"


def _load_configuration():
    config_dir = os.environ.get("XDG_CONFIG_HOME", "~/.config")
    config_filename = os.path.expanduser(os.path.join(config_dir, "lavacli.yaml"))

    try:
        with open(config_filename, encoding="utf-8") as f_conf:
            data = safe_yaml.load(f_conf.read())
            if not isinstance(data, dict):
                raise ConfigurationError("Invalid configuration file")
            return data
    except (FileNotFoundError, KeyError, TypeError):
        return {}


def _save_configuration(config):
    config_dir = os.environ.get("XDG_CONFIG_HOME", "~/.config")
    expanded_config_dir = os.path.expanduser(config_dir)
    config_filename = os.path.expanduser(os.path.join(config_dir, "lavacli.yaml"))

    if not os.path.exists(expanded_config_dir):
        os.makedirs(expanded_config_dir)

    with open(config_filename, "w", encoding="utf-8") as f_conf:
        flow_yaml.dump(config, f_conf)


def handle_add(_, options):
    config = _load_configuration()
    config[options.id] = {"uri": options.uri}
    if options.proxy:
        config[options.id]["proxy"] = options.proxy
    if options.username:
        config[options.id]["username"] = options.username
    if options.token:
        config[options.id]["token"] = options.token

    _save_configuration(config)
    return 0


def handle_delete(_, options):
    config = _load_configuration()
    try:
        del config[options.id]
    except KeyError:
        print("Unknown identity '%s'" % options.id, file=sys.stderr)
        return 1
    _save_configuration(config)
    return 0


def handle_list(_, __):
    config = _load_configuration()
    print("Identities:")
    for identity in sorted(config.keys()):
        print("* %s" % identity)
    return 0


def handle_show(_, options):
    config = _load_configuration()
    try:
        flow_yaml.dump(config[options.id], sys.stdout)
        return 0
    except KeyError:
        print("Unknown identity '%s'" % options.id, file=sys.stderr)
        return 1


def handle(proxy, options, _):
    handlers = {
        "add": handle_add,
        "delete": handle_delete,
        "list": handle_list,
        "show": handle_show,
    }
    return handlers[options.sub_sub_command](proxy, options)



#!/usr/bin/python3
# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import sys

from . import main

if __name__ == "__main__":
    sys.argv[0] = "lavacli"
    sys.exit(main())



#!/usr/bin/python3
# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

from setuptools import setup

# grab metadata without importing the module
metadata = {}
with open("lavacli/__about__.py", encoding="utf-8") as fp:
    exec(fp.read(), metadata)

# Setup the package
setup(
    name="lavacli",
    version=metadata["__version__"],
    description=metadata["__description__"],
    author=metadata["__author__"],
    author_email=metadata["__author_email__"],
    license=metadata["__license__"],
    url=metadata["__url__"],
    project_urls={
        "Bug Tracker": "https://gitlab.com/lava/lavacli/issues/",
        "Source Code": "https://gitlab.com/lava/lavacli/",
    },
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Communications",
        "Topic :: Software Development :: Testing",
        "Topic :: System :: Networking",
    ],
    packages=["lavacli", "lavacli.commands"],
    entry_points={"console_scripts": ["lavacli = lavacli:main"]},
    install_requires=["aiohttp", "jinja2", "requests", "ruamel.yaml", "voluptuous"],
    setup_requires=[],
    tests_require=["pytest", "pytest-runner"],
    zip_safe=True,
)



# vim: set ts=4

# Copyright 2023-present Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys

from lavacli.utils import safe_yaml


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    if version < (2023, 3):
        return

    # "add"
    groups_add = sub.add_parser("add", help="add a group")
    groups_add.add_argument("name", help="name")

    # "delete"
    groups_del = sub.add_parser("delete", help="delete a group")
    groups_del.add_argument("name", help="name")

    # "list"
    groups_list = sub.add_parser("list", help="list groups")
    out_format = groups_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        default=None,
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "perms"
    groups_perms = sub.add_parser("perms", help="manage user permissions")
    perms_sub = groups_perms.add_subparsers(
        dest="sub_sub_sub_command", help="Sub commands"
    )
    perms_sub.required = True
    perms_add = perms_sub.add_parser("add", help="add a permission")
    perms_add.add_argument("name", help="group name")
    perms_add.add_argument("app", help="application")
    perms_add.add_argument("model", help="model")
    perms_add.add_argument("codename", help="codename")

    perms_del = perms_sub.add_parser("delete", help="remove a permssion")
    perms_del.add_argument("name", help="group name")
    perms_del.add_argument("app", help="application")
    perms_del.add_argument("model", help="model")
    perms_del.add_argument("codename", help="codename")

    perms_list = perms_sub.add_parser("list", help="list permissions for a user")
    perms_list.add_argument("name", help="group name")
    out_format = perms_list.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        help="print as yaml",
    )

    # "show"
    groups_show = sub.add_parser("show", help="show group details")
    groups_show.add_argument("name", help="name")
    out_format = groups_show.add_mutually_exclusive_group()
    out_format.add_argument(
        "--json",
        dest="output_format",
        default=None,
        action="store_const",
        const="json",
        help="print as json",
    )
    out_format.add_argument(
        "--yaml",
        dest="output_format",
        action="store_const",
        const="yaml",
        default=None,
        help="print as yaml",
    )


def help_string():
    return "manage groups"


def handle_add(proxy, options, _):
    proxy.auth.groups.add(
        options.name,
    )
    return 0


def handle_delete(proxy, options, _):
    proxy.auth.groups.delete(options.name)
    return 0


def handle_list(proxy, options, _):
    groups = proxy.auth.groups.list()
    if options.output_format == "json":
        print(json.dumps(groups))
    elif options.output_format == "yaml":
        safe_yaml.dump(groups, sys.stdout)
    else:
        print("Groups:")
        for g in groups:
            print("* %s" % g)
    return 0


def handle_perms(proxy, options, _):
    if options.sub_sub_sub_command == "add":
        proxy.auth.groups.perms.add(
            options.name, options.app, options.model, options.codename
        )
    elif options.sub_sub_sub_command == "delete":
        proxy.auth.groups.perms.delete(
            options.name, options.app, options.model, options.codename
        )
    else:
        perms = proxy.auth.groups.perms.list(options.name)
        if options.output_format == "json":
            print(json.dumps(perms))
        elif options.output_format == "yaml":
            safe_yaml.dump(perms, sys.stdout)
        else:
            print("Permissions:")
            for perm in perms:
                print("* %s.%s.%s" % (perm["app"], perm["model"], perm["codename"]))
    return 0


def handle_show(proxy, options, config):
    group = proxy.auth.groups.show(options.name)

    if options.output_format == "json":
        print(json.dumps(group))
    elif options.output_format == "yaml":
        safe_yaml.dump(group, sys.stdout)
    else:
        print("name       : %s (%d)" % (group["name"], group["id"]))
        print("permissions: %s" % ", ".join(group["permissions"]))
        print("users      : %s" % ", ".join(group["users"]))


def handle(proxy, options, config):
    handlers = {
        "add": handle_add,
        "delete": handle_delete,
        "list": handle_list,
        "perms": handle_perms,
        "show": handle_show,
    }
    return handlers[options.sub_sub_command](proxy, options, config)
